!
! A MODULE for linear propagation of errors
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 20170527 $
!                               

module aderrors

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use constants
  use mod_linked_list
  use modBDIO
  
  IMPLICIT NONE

  private

  type :: uwname
     integer :: id
     character (len=:), allocatable :: name
  end type uwname

  type :: uwreal
     real (kind=DP), private :: mean, err = 0.0_DP, derr = 0.0_DP
     integer, private        :: nid=-1

     integer, allocatable, private        :: id(:)
     real (kind=DP), allocatable, private :: ti(:), dti(:)
     real (kind=DP), allocatable, private :: data(:), texp(:), &
          s(:), var(:), dsig(:)
     real (kind=DP), allocatable, private :: gamm(:), drho(:)
     integer,        allocatable, private :: nrep(:), tmax(:), &
          ivrep(:), nd(:), iw(:)

     real (kind=DP)   :: dlt(3) = 0.0_DP
   contains
     procedure           :: init => uwinit
     procedure           :: set_replica, set_id, set_texp, set_stau, set_dsig
     procedure, pass (p) :: get_offset, get_vrep_offset, get_cf_offset
     procedure           :: uwerr, error_src
     procedure           :: error_name => uwerror_name
     procedure           :: write_bdio, read_bdio
     procedure           :: print_hist
     procedure           :: neid, eid, taui, dtaui, window, value, error, derror
  end type uwreal

  real (kind=DP),     save :: DEFAULT_S_ULLI = 4.0_DP, DEFAULT_DSIG = 1.5_DP
  type (linked_list), save :: uwlst
  type (uwname),      save :: uwn


  integer, parameter, public   :: UW_CF_MODE = 0, UW_D_MODE = 1, UW_V_MODE = 2  
  integer,        save :: uwid_count = -2, imode = UW_CF_MODE
  real (kind=DP), save :: err_factor = 1.0_DP
  logical,        save :: light_mode = .false.
  
  public :: uwreal, aderr_new_id, aderr_id_exist, aderr_list_ids
  public :: aderr_dfactor, aderr_light_mode

  interface
     module function uwadd_cr(a,r)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwadd_cr
     end function
     module function uwadd_cl(r,a)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwadd_cl
     end function
     module function uwadd_t(a,b)
       type (uwreal), intent (in) :: a, b
       type (uwreal) :: uwadd_t
     end function
     module function uwneg(a)
       type (uwreal), intent (in)  :: a
       type (uwreal) :: uwneg
     end function
     module function uwsub_cr(a,r)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwsub_cr
     end function
     module function uwsub_cl(r,a)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwsub_cl
     end function
     module function uwsub_t(a,b)
       type (uwreal), intent (in) :: a, b
       type (uwreal) :: uwsub_t
     end function
     module function uwmul_cr(a,r)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwmul_cr
     end function
     module function uwmul_cl(r,a)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwmul_cl
     end function
     module function uwmul_t(a,b)
       type (uwreal), intent (in) :: a, b
       type (uwreal) :: uwmul_t
     end function
     module function uwdiv_cr(a,r)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwdiv_cr
     end function
     module function uwdiv_cl(r,a)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
       type (uwreal) :: uwdiv_cl
     end function
     module function uwdiv_t(a,b)
       type (uwreal), intent (in) :: a, b
       type (uwreal) :: uwdiv_t
     end function
     module function uwpow_in(a,n)
       type (uwreal), intent (in)  :: a
       integer, intent (in)        :: n
       type (uwreal) :: uwpow_in
     end function
     module function uwpow_re(a,d)
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: d
       type (uwreal) :: uwpow_re
     end function
     module function uwpow_t(a,b)
       type (uwreal), intent (in)  :: a, b
       type (uwreal) :: uwpow_t
     end function

     module function less_t(a,b)
       logical :: less_t
       type (uwreal), intent (in) :: a, b
     end function less_t
     module function less_dr(a,b)
       logical :: less_dr
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function less_dr
     module function less_dl(b,a)
       logical :: less_dl
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function less_dl
     module function leq_t(a,b)
       logical :: leq_t
       type (uwreal), intent (in) :: a, b
     end function leq_t
     module function leq_dr(a,b)
       logical :: leq_dr
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function leq_dr
     module function leq_dl(b,a)
       logical :: leq_dl
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function leq_dl
     module function gt_t(a,b)
       logical :: gt_t
       type (uwreal), intent (in) :: a, b
     end function gt_t
     module function gt_dr(a,b)
       logical :: gt_dr
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function gt_dr
     module function gt_dl(b,a)
       logical :: gt_dl
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function gt_dl
     module function geq_t(a,b)
       logical :: geq_t
       type (uwreal), intent (in) :: a, b
     end function geq_t
     module function geq_dr(a,b)
       logical :: geq_dr
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function geq_dr
     module function geq_dl(b,a)
       logical :: geq_dl
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: b
     end function geq_dl

     
     module subroutine uwdobs_f(x, a, fn)
       type (uwreal), intent (inout) :: x
       type (uwreal), intent (in) :: a(:)
       interface 
          function fn(z)
            USE numtypes
            real (kind=DP) :: fn
            real (kind=DP), intent (in) :: z(:)
          end function fn
       end interface
     end subroutine
     module subroutine uwdobs_s(x, a, sn)
       type (uwreal), intent (inout) :: x(:)
       type (uwreal), intent (in) :: a(:)
       interface 
          subroutine sn(r,z)
            USE numtypes
            real (kind=DP), intent (inout) :: r(:)
            real (kind=DP), intent (in)  :: z(:)
          end subroutine sn
       end interface
     end subroutine 

     module subroutine addobs_f(x, a, fn)
       type (uwreal), intent (inout) :: x
       type (uwreal), intent (in)    :: a(:)

       interface 
          function fn(z)
            USE numtypes
            import :: uwreal
            type (uwreal) :: fn
            type (uwreal), intent (in) :: z(:)
          end function fn
       end interface
     end subroutine
  
     module subroutine addobs_s(x, a, sn)
       type (uwreal), intent (inout) :: x(:)
       type (uwreal), intent (in)    :: a(:)

       interface 
          subroutine sn(r,z)
            USE numtypes
            import :: uwreal
            type (uwreal), intent (inout) :: r(:)
            type (uwreal), intent (in)    :: z(:)
          end subroutine sn
       end interface
     end subroutine

     module subroutine addobs_d(x, a, der, mns)
       type (uwreal), intent  (inout) :: x(:)
       type (uwreal), intent  (in)    :: a(:)
       real (kind=DP), intent (in)    :: der(size(x),size(a)), mns(size(x))
     end subroutine

     module subroutine addobs_ds(x, a, der, mns)
       type (uwreal), intent  (inout) :: x
       type (uwreal), intent  (in)    :: a(:)
       real (kind=DP), intent (in)    :: der(:), mns
     end subroutine
  end interface

  interface assignment (=)
     module procedure uwequal, uwequal_t, uwequal_s
  end interface assignment (=)
  interface operator(+)
     module procedure uwadd_cr, uwadd_cl, uwadd_t
  end interface operator(+)
  interface operator(-)
     module procedure uwneg, uwsub_cr, uwsub_cl, uwsub_t
  end interface operator(-)
  interface operator(*)
     module procedure uwmul_cr, uwmul_cl, uwmul_t
  end interface operator(*)
  interface operator(/)
     module procedure uwdiv_cr, uwdiv_cl, uwdiv_t
  end interface operator(/)
  interface operator(**)
     module procedure uwpow_in, uwpow_re, uwpow_t
  end interface operator(**)
  interface operator(.lt.)
     module procedure less_t, less_dr, less_dl
  end interface operator(.lt.)
  interface operator(.le.)
     module procedure leq_t, leq_dr, leq_dl
  end interface operator(.le.)
  interface operator(.gt.)
     module procedure gt_t, gt_dr, gt_dl
  end interface operator(.gt.)
  interface operator(.ge.)
     module procedure geq_t, geq_dr, geq_dl
  end interface operator(.ge.)
  
  public :: operator(+), operator(-), operator(*), &
       operator(/), operator(**), assignment(=)
  public :: operator(.lt.), operator(.gt.), operator(.le.), &
       operator(.ge.)

  interface uwdobs
     module procedure uwdobs_f, uwdobs_s
  end interface uwdobs
  interface addobs
     module procedure addobs_f, addobs_s, addobs_d, addobs_ds
  end interface addobs

  public :: uwdobs, addobs
  
  interface
     module subroutine uwerr(p, ids_texp)
       class (uwreal), intent (inout) :: p
       integer, intent (in), optional :: ids_texp(:)
     end subroutine
  end interface

  interface
     module elemental function uwsin(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwsin
     module elemental function uwcos(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwcos
     module elemental function uwtan(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwtan
     module elemental function uwasin(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwasin
     module elemental function uwacos(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwacos
     module elemental function uwatan(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwatan
     module elemental function uwsinh(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwsinh
     module elemental function uwcosh(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwcosh
     module elemental function uwtanh(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwtanh
     module elemental function uwsqrt(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwsqrt
     module elemental function uwlog(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwlog
     module elemental function uwlog10(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwlog10
     module elemental function uwexp(a) result (p)
       type (uwreal), intent (in) :: a
       type (uwreal) :: p
     end function uwexp
  end interface

  interface sin
     module procedure uwsin
  end interface sin
  interface cos
     module procedure uwcos
  end interface cos
  interface tan
     module procedure uwtan
  end interface tan
  
  interface asin
     module procedure uwasin
  end interface asin
  interface acos
     module procedure uwacos
  end interface acos
  interface atan
     module procedure uwatan
  end interface atan
  
  interface sinh
     module procedure uwsinh
  end interface sinh
  interface cosh
     module procedure uwcosh
  end interface cosh
  interface tanh
     module procedure uwtanh
  end interface tanh
  
  interface sqrt
     module procedure uwsqrt
  end interface sqrt
  interface log
     module procedure uwlog
  end interface log
  interface log10
     module procedure uwlog10
  end interface log10
  interface exp
     module procedure uwexp
  end interface exp

  public :: sin, cos, tan, asin, acos, atan, sinh, cosh, tanh, &
       sqrt, log, log10, exp

  
  interface
     module subroutine write_bdio(p, fb, iu)
       class (uwreal), intent (inout) :: p
       type (BDIO), intent (inout) :: fb
       integer, intent (in) :: iu
     end subroutine
     module subroutine read_bdio(p, fb)
       class (uwreal), intent (inout) :: p
       type (BDIO), intent (inout) :: fb
     end subroutine

     module function neid(p) result(x)
       class (uwreal), intent (in) :: p
       integer                     :: x
     end function neid
     module function eid(p,i) result(x)
       class (uwreal), intent (in) :: p
       integer, intent (in)        :: i 
       integer                     :: x
     end function eid
     module function dtaui(p,i) result(x)
       class (uwreal), intent (in) :: p
       integer, intent (in)        :: i 
       real (kind=DP)              :: x
     end function
     module function taui(p,i) result(x)
       class (uwreal), intent (in) :: p
       integer, intent (in)        :: i 
       real (kind=DP)              :: x
     end function
     module function window(p,i) result(x)
       class (uwreal), intent (in) :: p
       integer, intent (in)        :: i 
       integer                     :: x
     end function window
     module elemental function value(p) result(x)
       class (uwreal), intent (in) :: p
       real (kind=DP)              :: x
     end function
     module function error(p)
       class (uwreal), intent (in) :: p
       real (kind=DP)              :: error
     end function
     module function derror(p)
       class (uwreal), intent (in) :: p
       real (kind=DP)              :: derror
     end function
  end interface

  interface
     module subroutine aderr_cov(mat,p)
       type (uwreal),  intent (in)    :: p(:)
       real (kind=DP), intent (inout) :: mat(:,:)
     end subroutine
  end interface

  interface 
     module subroutine uwerr_grad_f(grad, f, x)
       type (uwreal), intent (in)    :: x(:)
       real (kind=DP), intent (inout) :: grad(size(x))

       interface 
          function f(x)
            import :: uwreal
            type (uwreal), intent (in) :: x(:)
            type (uwreal) :: f
          end function f
       end interface 
     end subroutine
     module subroutine uwerr_grad_s(grad, s, x)
       type (uwreal), intent (in)    :: x(:)
       real (kind=DP), intent (inout) :: grad(:,:)
       
       interface 
          subroutine s(f, x)
            import :: uwreal
            type (uwreal), intent (in)    :: x(:)
            type (uwreal), intent (inout) :: f(:)
          end subroutine s
       end interface
     end subroutine

     module subroutine uwerr_hess_s(hess, s, x)
       type (uwreal),  intent (in)    :: x(:)
       real (kind=DP), intent (inout) :: hess(:,:,:)
       
       interface 
          subroutine s(f, x)
            import :: uwreal
            type (uwreal), intent (in)    :: x(:)
            type (uwreal), intent (inout) :: f(:)
          end subroutine s
       end interface
     end subroutine

     module subroutine uwerr_hess_f(hess, f, x)
       type (uwreal),  intent (in) :: x(:)
       real (kind=DP), intent (inout) :: hess(size(x),size(x))
       
       interface 
          function f(x)
            import :: uwreal
            type (uwreal), intent (in) :: x(:)
            type (uwreal) :: f
          end function f
       end interface
     end subroutine
  end interface
    
  interface aderr_grad
     module procedure uwerr_grad_f, uwerr_grad_s
  end interface aderr_grad
  interface aderr_hess
     module procedure uwerr_hess_f, uwerr_hess_s
  end interface aderr_hess
  
  public :: aderr_cov, aderr_grad, aderr_hess

  interface
     module function dadd_dt(a,b) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in) :: a, b
     end function
     module function dadd_dr(a,r) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dadd_dl(r,a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dneg(a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
     end function
     module function dsub_dt(a,b) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in) :: a, b
     end function
     module function dsub_dr(a,r) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dsub_dl(r,a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dmul_dt(a,b) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in) :: a, b
     end function
     module function dmul_dr(a,r) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dmul_dl(r,a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function ddiv_dt(b,a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in) :: a, b
     end function
     module function ddiv_dr(a,r) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function ddiv_dl(r,a) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in)  :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dpow_in(a,n) result (res)
       type (uwreal) :: res
       type (uwreal), intent (in) :: a
       integer, intent (in)       :: n
     end function
     module function dpow_re(a,r) result (res)
       type (uwreal) :: res
       type (uwreal),  intent (in) :: a
       real (kind=DP), intent (in) :: r
     end function
     module function dpow_dt(a,b) result (res)
       type (uwreal) :: res
       type (uwreal),  intent (in) :: a, b
     end function
  end interface

  public :: aderr_set_mode, aderr_get_mode, adset_default_dsig, &
       adset_default_stau, aderr_pobs
  
contains

! *****************************
! *
  subroutine adset_default_stau(t)
! *
! *****************************
    real (kind=DP), intent (in)  :: t

    DEFAULT_S_ULLI = t
    
    return
  end subroutine adset_default_stau
  
! *****************************
! *
  subroutine adset_default_dsig(ds)
! *
! *****************************
    real (kind=DP), intent (in)  :: ds

    DEFAULT_DSIG = ds
    
    return
  end subroutine adset_default_dsig
  
! *****************************
! *
  subroutine aderr_set_mode(n)
! *
! *****************************
    integer, intent (in) :: n

    if ( (n.lt.0).or.(n.gt.2) ) then
       call module_error('[uwerr_set_mode]', &
            'mode out of range [0:2]')
    end if
    imode = n
    
    return
  end subroutine aderr_set_mode
  
! *****************************
! *
  function aderr_get_mode() result (n)
! *
! *****************************
    integer :: n

    n = imode
    
    return
  end function aderr_get_mode
  
! *****************************
! *
  subroutine aderr_light_mode()
! *
! *****************************

    light_mode = .true.
    
    return
  end subroutine aderr_light_mode
  
! ******************************
! *
  subroutine aderr_dfactor(val)
    real (kind=DP), intent(in)  :: val
    
    err_factor = val
    
    return
  end subroutine aderr_dfactor
  
! ****************************************
! *
  pure subroutine uwfree(p)
! *
! ****************************************
    class (uwreal), intent (inout) :: p

    if (allocated(p%nd   )) deallocate(p%nd   )
    if (allocated(p%nrep )) deallocate(p%nrep )
    if (allocated(p%tmax )) deallocate(p%tmax )
    if (allocated(p%texp )) deallocate(p%texp )
    if (allocated(p%iw   )) deallocate(p%iw   )
    if (allocated(p%id   )) deallocate(p%id   )
    if (allocated(p%s    )) deallocate(p%s    )
    if (allocated(p%dsig )) deallocate(p%dsig )
    if (allocated(p%ivrep)) deallocate(p%ivrep)
    if (allocated(p%data )) deallocate(p%data )
    if (allocated(p%gamm )) deallocate(p%gamm )
    if (allocated(p%drho )) deallocate(p%drho )
    if (allocated(p%var  )) deallocate(p%var  )
    if (allocated(p%ti   )) deallocate(p%ti   )
    if (allocated(p%dti  )) deallocate(p%dti  )
    p%mean = 0.0_DP
    p%err  = 0.0_DP
    p%derr = 0.0_DP
    p%dlt  = 0.0_DP
    
    return
  end subroutine uwfree
    
! ****************************************
! *
  pure subroutine uwinit(p, nid, nd, nrep, ivrep)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    integer, intent (in) :: nid, nd(:), nrep(:), ivrep(:)

    integer :: i, is, ie, n

    call uwfree(p)
    p%nid = nid
    allocate(p%nd(nid), p%nrep(nid), p%tmax(nid), p%var(nid), p%dsig(nid), &
         p%texp(nid), p%iw(nid), p%id(nid), p%s(nid), p%ti(nid), p%dti(nid))
    allocate(p%ivrep(size(ivrep)))
    
    p%nd    = nd(1:nid)
    p%nrep  = nrep(1:nid)
    p%id    = -1
    p%texp  = 0.0_DP
    p%s     = DEFAULT_S_ULLI
    p%ivrep = ivrep
    p%var   = 0.0_DP
    p%dsig  = DEFAULT_DSIG
    p%ti    = 0.0_DP
    p%dti   = 0.0_DP
    
    allocate(p%data(sum(p%nd)))

    do i = 1, p%nid
       call p%get_vrep_offset(is,ie,i)
       p%tmax(i) = maxval(p%ivrep(is:ie))/2
    end do
    n = sum(p%tmax)
    if (allocated(p%gamm)) deallocate(p%gamm)
    if (allocated(p%drho)) deallocate(p%drho)
    
    return
  end subroutine uwinit
    
! ****************************************
! *
  subroutine uwequal_t(p, o)
! *
! ****************************************
    type (uwreal), intent (inout) :: p
    type (uwreal), intent (in)    :: o

    if (imode.eq.UW_CF_MODE) then
       call p%init(o%nid, o%nd, o%nrep, o%ivrep)
       
       p%nrep  = o%nrep
       p%id    = o%id
       p%texp  = o%texp
       p%dsig  = o%dsig
       p%s     = o%s
       p%iw    = o%iw

       p%mean = o%mean
       p%data = o%data
       p%tmax = o%tmax
    else
       call uwfree(p)
       p%mean = o%mean
       p%dlt  = o%dlt
    end if
       
    return
  end subroutine uwequal_t

! ****************************************
! *
  recursive subroutine uwequal(p, d)
! *
! ****************************************
    type (uwreal), intent (inout) :: p
    real (kind=DP), intent (in)   :: d(:)

    integer       :: n
    
    n = size(d)
    if (imode.eq.UW_CF_MODE) then
       if (n.eq.1) call module_error('[uwequal]', &
            'uwreal data type requires at least 2 data values')

       if (n.eq.2) then
          call p%init(1, [1], [1], [1])

          p%mean     = d(1)
          p%data     = d(2)
          p%ivrep(1) = p%nd(1)

          return
       else
          call p%init(1, [n], [1], [n])

          p%mean = sum(d)/real(p%nd(1),kind=DP)
          p%data = d-p%mean
       end if

       if (light_mode) then
          call p%uwerr()
          p = [p%mean, p%err]
       end if
    else
       call uwfree(p)
       if (n.ne.4) call module_error('[uwequal]', &
            'uwreal data type in D_MODE requires 4 data values')

       p%mean     = d(1)
       p%dlt(1:3) = d(2:4)
    end if
           
    return
  end subroutine uwequal
  
! ****************************************
! *
  subroutine uwequal_s(p, d)
! *
! ****************************************
    type (uwreal), intent (inout) :: p
    real (kind=DP), intent (in)   :: d

    
    if (imode.eq.UW_CF_MODE) then
       call p%init(1, [1], [1], [1])
       p%mean = d
       p%data = 0.0_DP
       p%ivrep(1) = p%nd(1)
    else
       call uwfree(p)
       p%mean     = d
       p%dlt(1:3) = 0.0_DP
    end if
       
    return
  end subroutine uwequal_s
  
! ****************************************
! *
  subroutine get_offset(is, ie, p, id, rp)
! *
! ****************************************
    integer, intent (out)       :: is, ie
    class (uwreal), intent (in) :: p
    integer, intent (in) :: id, rp

    integer :: iro

    if (rp.eq.0) then
       is  = sum(p%nd(1:id-1)) + 1
       ie  = is + p%nd(id) - 1
    else
       iro = sum(p%nrep(1:id-1))
       is  = sum(p%nd(1:id-1)) + sum(p%ivrep(iro+1:iro+rp-1)) + 1
       ie  = is + p%ivrep(iro+rp) - 1
    end if
    
    return
  end subroutine get_offset
    
! ****************************************
! *
  pure subroutine get_vrep_offset(is, ie, p, id)
! *
! ****************************************
    integer, intent (out)       :: is, ie
    class (uwreal), intent (in) :: p
    integer, intent (in) :: id

    is  = sum(p%nrep(1:id-1)) + 1
    ie  = is + p%nrep(id) - 1
    
    return
  end subroutine get_vrep_offset
    
! ****************************************
! *
  subroutine get_cf_offset(is, ie, p, id)
! *
! ****************************************
    integer, intent (out)       :: is, ie
    class (uwreal), intent (in) :: p
    integer, intent (in) :: id

    is  = sum(p%tmax(1:id-1))
    ie  = is + p%tmax(id) - 1
    
    return
  end subroutine get_cf_offset

! ****************************************
! *
  subroutine aderr_list_ids(ifp)
! *
! ****************************************
    integer, intent (in), optional :: ifp
    
    integer :: ifn
    class (*), pointer :: dt
    logical :: ok

    if (present(ifp)) then
       ifn = ifp
    else
       ifn = output_unit
    end if
    
    ok = uwlst%scan(0)
    if (.not.ok) return
    do
       call uwlst%getdata(dt)
       select type (dt)
       type is (uwname)
          write(ifn,'(3X,1A,2X,1I0,2A)')' ID:', dt%id, ' NAME:', dt%name
       end select
       if (.not.uwlst%scan()) return
    end do
    

    return
  end subroutine aderr_list_ids
    
! ****************************************
! *
  function aderr_new_id(name, id)
! *
! ****************************************
    character (len=*), intent (in) :: name
    integer, intent (in), optional :: id
    integer :: aderr_new_id
    
    if (present(id)) then
       aderr_new_id = id
    else
       aderr_new_id = uwid_count
       uwid_count   = uwid_count - 1
    end if
    uwn%id   = aderr_new_id
    uwn%name = name
    call uwlst%add(uwn)
    
    return
  end function aderr_new_id
  
! ****************************************
! *
  function aderr_id_exist(id) result (ex)
! *
! ****************************************
    integer, intent (in) :: id
    logical :: ex
    
    class (*), pointer :: dt
    logical :: ok
   
    ex = .false.
    ok = uwlst%scan(0)
    if (.not.ok) then
       ex = .false.
       return
    end if
    do
       call uwlst%getdata(dt)
       select type (dt)
       type is (uwname)
          if (dt%id==id) then
             ex = .true.
             return
          end if
       end select
       if (.not.uwlst%scan()) return
    end do
    
    return
  end function aderr_id_exist
    
! ****************************************
! *
  function uwerror_name(p, id)
! *
! ****************************************
    class (uwreal), intent (in)    :: p
    integer, intent (in)           :: id
    character (len=:), allocatable :: uwerror_name
    
    class (*), pointer :: dt
    logical :: ok
    
    ok = uwlst%scan(0)
    if (.not.ok) call module_error('[error_name]', &
         'Error reaching beginning of linked list')
    do
       call uwlst%getdata(dt)
       select type (dt)
       type is (uwname)
          if (dt%id==p%id(id)) then
             uwerror_name = dt%name
             exit
          end if
       end select
       if (.not.uwlst%scan()) call module_error('[error_name]', &
         'ID not present in database')
    end do
    
    return
  end function uwerror_name
    
! ****************************************
! *
  function error_src(p, i)
! *
! ****************************************
    class (uwreal), intent (in) :: p
    integer, intent (in)        :: i 
    real (kind=DP) :: error_src

    error_src = p%var(i)/p%err**2
       
    return
  end function error_src

! ****************************************
! *
  subroutine set_texp(p, tx)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    real (kind=DP), intent (in)    :: tx

    if (p%nid.ne.1) call module_error('[set_texp]', &
         'tau_exp can only be set on primary observables')

    p%texp(1) = tx

    return
  end subroutine set_texp
    
! ****************************************
! *
  subroutine set_stau(p, tx, id)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    real (kind=DP), intent (in)    :: tx
    integer, intent (in), optional :: id
    
    integer :: idx, i
    logical :: found
    
    found = .false.
    if (present(id)) then
       idx = -1
       do i = 1, p%neid()
          if (p%eid(i).eq.id) then
             idx = i
             found = .true.
             exit
          end if
       end do
    else
       idx = 1
       found = .true.
    end if
       
    if (.not.found) call module_error('[set_stau]', &
         'id not contributing to this observable')

    p%s(idx) = tx

    return
  end subroutine set_stau
    
! ****************************************
! *
  subroutine set_dsig(p, ds, id)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    real (kind=DP), intent (in)    :: ds
    integer, intent (in), optional :: id

    integer :: idx, i
    logical :: found
    
    found = .false.
    if (present(id)) then
       idx = -1
       do i = 1, p%neid()
          if (p%eid(i).eq.id) then
             idx = i
             found = .true.
             exit
          end if
       end do
    else
       idx = 1
       found = .true.
    end if

    if (.not.found) call module_error('[set_dsig]', &
         'id not contributing to this observable')
    p%dsig(idx) = ds

    return
  end subroutine set_dsig
    
! ****************************************
! *
  subroutine set_id(p, id)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    integer, intent (in)           :: id

    if (p%nid.ne.1) call module_error('[set_id]', &
         'ID can only be set on primary observables')

    p%id(1) = id
    
    return
  end subroutine set_id
    
! ****************************************
! *
  subroutine set_replica(p, vrep)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    integer, intent (in), optional :: vrep(:)

    integer :: is, ie, irs, ire, iacum
    
    if (light_mode) return
    if (p%nid.ne.1) call module_error('[set_replica]', &
         'Replica can only be set on primary observables')
    
    if (allocated(p%ivrep)) deallocate(p%ivrep)
    if (present(vrep)) then
       p%nrep(1) = size(vrep)
       allocate(p%ivrep(p%nrep(1)))
       
       p%ivrep(:) = vrep(:)
       p%tmax(1)  = int(maxval(p%ivrep)/2)
       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       allocate(p%gamm(0:p%tmax(1)-1), p%drho(0:p%tmax(1)-1))
       
       is = 1
       ie = p%nd(1)
       irs = 1
       ire = p%nrep(1)
       iacum = sum(p%ivrep(irs:ire))-(ie-is+1)
       if (iacum.ne.0) call module_error('[set_replica]', &
            'Measurements do not match sum over replica')
    else
       p%nrep(1) = 1
       allocate(p%ivrep(p%nrep(1)))
       
       p%ivrep(1) = p%nd(1)
       p%tmax(1)  = int(p%ivrep(1)/2)
    end if
       
    return
  end subroutine set_replica

! *****************************
! *
  subroutine print_hist(p, iifn)
! *
! *****************************
    class (uwreal), intent (in)    :: p
    integer, intent (in), optional :: iifn

    integer        :: ifn, i, k, nv, ist, ioff
    real (kind=DP) :: acum, tti, ttex, dtti, ersq
    
    if (present(iifn)) then
       ifn = iifn
    else
       ifn = output_unit
    end if

    write(ifn,*)'####'
    write(ifn,'(1X,1A,1ES25.15," +/- ",1ES10.4)')'# OBSERVABLE: ', p%mean, p%err
    write(ifn,'(1X,1A,1I0,2A)')'####   ', p%nid, ' Ensembles contributing ', '####'
    do i = 1, p%nid
       nv = sum(p%nrep(1:i-1))
       write(ifn,'(1X,1A,1I4,1A)')'#    (*) Ensemble ID: ', p%id(i)
       write(ifn,'(1X,1A,1F6.2,1A)')'#       - Contribution to error: ', p%error_src(i)*100, ' %'
       write(ifn,'(1X,1A,1I0,1A,1I0,1A)')'#       - ', p%nd(i), ' measurements in ', p%nrep(i), ' replica'
       write(ifn,'(1X,1A,1000(I0,2X))')'#       - Replica vector:   ', p%ivrep(nv+1:nv+p%nrep(i))
       write(ifn,'(1X,1A,1F8.1)')'#       - Parameter S:      ', p%s(i)
       write(ifn,'(1X,1A,1F8.1)')'#       - Parameter tauexp: ', p%texp(i)
       write(ifn,'(1X,1A,1F8.1)')'#       - Parameter dsig:   ', p%dsig(i)
       write(ifn,'(1X,1A,1I0)')'#       - Window:           ', p%iw(i)
       write(ifn,'(1X,1A,1F6.2,1A,1F6.2)')'#       - Tauint:           ', p%ti(i), ' +/- ', p%dti(i)
    end do
    write(ifn,*)

    ioff = 0
    do i = 1, p%nid
       ist = sum(p%tmax(1:i-1))
       write(ifn,'(6ES23.15)',advance="NO")p%mean

       if (p%nd(i).gt.1) then
          write(ifn,'(6ES23.15)')p%gamm(ist), 1.0, p%drho(ist), 0.5, 0.0
          acum = 0.50_DP
          do k = 1, p%nd(i)
             write(ifn,'(6ES23.15)',advance="NO")p%data(ioff+k) + p%mean
             if (k.lt.p%tmax(i)-1) then
                acum = acum + p%gamm(ist+k)/p%gamm(ist) 
                tti  = acum 
                ttex = p%texp(i)*p%gamm(ist+k)/p%gamm(ist)
                ersq = (real(4*k+2,kind=DP)-2.0_DP*tti)/real(p%nd(i),kind=DP)
                dtti = sqrt(tti**2*ersq + p%texp(i)**2*p%drho(ist+k)**2)
                write(ifn,'(6ES23.15)',advance="NO")p%gamm(ist+k), p%gamm(ist+k)/p%gamm(ist), p%drho(ist+k), tti+ttex, dtti
             else
                write(ifn,'(5A23)',advance="NO")'NaN','NaN','NaN','NaN','NaN'
             end if
             write(ifn,*)
          end do
       else
          write(ifn,'(5A23)',advance="NO")'NaN','NaN','NaN','NaN','NaN'
       end if
       
       write(ifn,*)
       write(ifn,*)
       write(ifn,*)
       ioff = ioff + p%nd(i)
    end do
    

    return
  end subroutine print_hist

! *****************************
! *
  function aderr_pobs(data, id_msm, ivrep, nmsm) result (p)
! *
! *****************************
    type (uwreal) :: p
    real (kind=DP), intent (in)   :: data(:)
    integer, intent (in)          :: id_msm(:), ivrep(:), nmsm

    integer :: i

    if (size(data).ne.size(id_msm)) call module_error('[aderr_pobs]',&
         ' Data and trajectories do not match')
    
    call p%init(1, [nmsm], [size(ivrep)], ivrep)

    p%mean = sum(data)/real(size(data),kind=DP)
    p%data = 0.0_DP
    do i = 1, size(id_msm)
       p%data(id_msm(i)) = data(i) - p%mean
    end do
    p%data = real(nmsm,kind=DP)/real(size(id_msm)) * p%data
    
    return
  end function aderr_pobs

! ********************************
! *
    Subroutine module_error(routine, msg)
! *
! ********************************

      Character (len=*), Intent (in) :: routine, msg

      Write(error_unit,*)'error '//Trim(routine)// ':'
      Write(error_unit,'(5X,1A)')Trim(msg)
      Write(error_unit,*)

      stop
    end subroutine module_error

end Module aderrors


! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:




