!
! MODULE uncertainties 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    uwerrors_darith.f90
! created: Thu Jul  5 00:59:53 2018
!                               

submodule (aderrors) aderrors_darith

  implicit none
  
contains

! ********************************
! *
  module function dadd_dt(a,b) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in) :: a, b
    
    res%mean = a%mean + b%mean
    res%dlt  = a%dlt + b%dlt
    
    return
  end function
  
! ********************************
! *
  module function dadd_dr(a,r) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%dlt  = a%dlt
    res%mean = a%mean + r
    
    Return
  end function
    
! ********************************
! *
  module function dadd_dl(r,a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%dlt  = a%dlt
    res%mean = a%mean + r
    
    Return
  end function
    
! ********************************
! *
  module function dneg(a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a

    res%dlt  = -a%dlt
    res%mean = -a%mean
    
    Return
  end function
    
! ********************************
! *
  module function dsub_dt(a,b) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in) :: a, b
    
    res%dlt  = a%dlt  - b%dlt
    res%mean = a%mean - b%mean
    
    return
  end function
  
! ********************************
! *
  module function dsub_dr(a,r) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%dlt  = a%dlt
    res%mean = a%mean - r
    
    Return
  end function
    
! ********************************
! *
  module function dsub_dl(r,a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%dlt  = -a%dlt
    res%mean = -a%mean + r
    
    Return
  end function
    
! ********************************
! *
  module function dmul_dt(a,b) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in) :: a, b
    
    res%mean = a%mean * b%mean
    res%dlt(1) = a%mean*b%dlt(1) + a%dlt(1)*b%mean
    res%dlt(2) = a%mean*b%dlt(2) + a%dlt(2)*b%mean
    res%dlt(3) = a%mean*b%dlt(3) + a%dlt(1)*b%dlt(2) + &
         a%dlt(2)*b%dlt(1) + a%dlt(3)*b%mean
    
    return
  end function
  
! ********************************
! *
  module function dmul_dr(a,r) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%mean = r*a%mean
    if (imode.eq.UW_D_MODE) res%dlt = r*a%dlt
    
    Return
  end function
    
! ********************************
! *
  module function dmul_dl(r,a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%mean = r*a%mean
    if (imode.eq.UW_D_MODE) res%dlt = r*a%dlt
    
    Return
  end function
    
! ********************************
! *
  module function ddiv_dt(b,a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in) :: a, b
    
    res%mean = b%mean/a%mean
    if (imode.eq.UW_D_MODE) then
       res%dlt(1) = b%dlt(1)/a%mean - a%dlt(1)*b%mean/a%mean**2
       res%dlt(2) = b%dlt(2)/a%mean - a%dlt(2)*b%mean/a%mean**2
       res%dlt(3) = b%dlt(3)/a%mean - a%dlt(1)*b%dlt(2)/a%mean**2 - &
            a%dlt(2)*b%dlt(1)/a%mean**2 + &
            b%mean*(2.0_DP*a%dlt(1)*a%dlt(2)/a%mean - a%dlt(3))/a%mean**2
    end if
    
    return
  end function
  
! ********************************
! *
  module function ddiv_dr(a,r) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%mean = a%mean/r
    if (imode.eq.UW_D_MODE) res%dlt  = a%dlt/r
    
    Return
  end function
    
! ********************************
! *
  module function ddiv_dl(r,a) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r

    res%mean = r/a%mean
    if (imode.eq.UW_D_MODE) then
       res%dlt(1) = - a%dlt(1)*r/a%mean**2
       res%dlt(2) = - a%dlt(2)*r/a%mean**2
       res%dlt(3) = r*(2.0_DP*a%dlt(1)*a%dlt(2)/a%mean - &
            a%dlt(3))/a%mean**2
    end if
       
    Return
  end function

! ********************************
! *
  module function dpow_in(a,n) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal), intent (in) :: a
    integer, intent (in)       :: n

    real (kind=DP) :: d, dd
        
    if (imode.eq.UW_D_MODE) then
       select case (n)
       case (0)
          d  = 0.0_DP
          dd = 0.0_DP
       case (1)
          d  = 1.0_DP
          dd = 0.0_DP
       case default
          d  = n*a%mean**(n-1)
          dd = real(n*(n-1),kind=DP)*a%mean**(n-2)
       end select
    end if

    res%mean = a%mean**n
    if (imode.eq.UW_D_MODE) then
       res%dlt(1) = d  * a%dlt(1)
       res%dlt(2) = d  * a%dlt(2)
       res%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
    end if
    
    return
  end function
    
! ********************************
! *
  module function dpow_re(a,r) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal),  intent (in) :: a
    real (kind=DP), intent (in) :: r

    real (kind=DP) :: d, dd
        
    res%mean = a%mean**r
    if (imode.eq.UW_D_MODE) then
       d  = r*a%mean**(r-1.0_DP)
       dd = r*(r-1.0_DP)*a%mean**(r-2.0_DP)

       res%dlt(1) = d  * a%dlt(1)
       res%dlt(2) = d  * a%dlt(2)
       res%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
    end if
    
    return
  end function

! ********************************
! *
  module function dpow_dt(a,b) result (res)
! *
! ********************************
    type (uwreal) :: res
    type (uwreal),  intent (in) :: a, b

    type (uwreal) :: lg
            
    if (imode.eq.UW_D_MODE) then
       lg  = b*log(a)
       res = exp(lg)
    else
       res%mean = a%mean**b%mean
    end if
    
    return
  end function

! ********************************
! *
  subroutine submodule_error(routine, msg)
! *
! ********************************

    character (len=*), intent (in) :: routine, msg

    write(error_unit,*)'error '//trim(routine)// ':'
    write(error_unit,'(5x,1a)')trim(msg)
    write(error_unit,*)

    stop
  end subroutine submodule_error

  
end submodule aderrors_darith
