!
! MODULE name 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    aderrors_io.f90
! created: Apr 30 11:44:28 2018
!                               

submodule (aderrors) aderrors_io

  implicit none
  
  
contains

! *****************************
! *
  module function neid(p) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    integer                     :: x

    x = p%nid

    return
  end function neid
    
! *****************************
! *
  module function eid(p,i) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    integer, intent (in)        :: i 
    integer                     :: x

    x = p%id(i)

    return
  end function eid
    
! *****************************
! *
  module function dtaui(p,i) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    integer, intent (in)        :: i 
    real (kind=DP)              :: x

    x = p%dti(i)

    return
  end function dtaui
    
! *****************************
! *
  module function taui(p,i) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    integer, intent (in)        :: i 
    real (kind=DP)              :: x

    x = p%ti(i)

    return
  end function taui
    
! *****************************
! *
  module function window(p,i) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    integer, intent (in)        :: i 
    integer                     :: x

    x = p%iw(i)

    return
  end function window
    
! *****************************
! *
  module elemental function value(p) result(x)
! *
! *****************************
    class (uwreal), intent (in) :: p
    real (kind=DP)              :: x

    x = p%mean

    return
  end function value
  
! *****************************
! *
  module function error(p)
! *
! *****************************
    class (uwreal), intent (in) :: p
    real (kind=DP)              :: error

    error = p%err

    return
  end function error
  
! *****************************
! *
  module function derror(p)
! *
! *****************************
    class (uwreal), intent (in) :: p
    real (kind=DP)              :: derror

    derror = p%derr

    return
  end function derror
  
! ****************************************
! *
  module subroutine write_bdio(p, fb, iu)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    type (BDIO), intent (inout) :: fb
    integer, intent (in) :: iu

    real (kind=DP) :: dfoo(1)
    integer :: ifoo(1)
    integer :: iw

    call BDIO_start_record(fb,BDIO_BIN_GENERIC,iu,.true.)
    ifoo(1) = p%nid
    dfoo(1) = p%mean
    iw = 0
    iw = iw + BDIO_write(fb,dfoo,    .true.)
    iw = iw + BDIO_write(fb,ifoo,    .true.)
    iw = iw + BDIO_write(fb,p%nd,    .true.)
    iw = iw + BDIO_write(fb,p%nrep,  .true.)
    iw = iw + BDIO_write(fb,p%ivrep, .true.)
    iw = iw + BDIO_write(fb,p%id,    .true.)
    iw = iw + BDIO_write(fb,p%tmax,  .true.)
    iw = iw + BDIO_write(fb,p%texp,  .true.)
    iw = iw + BDIO_write(fb,p%s,     .true.)
    iw = iw + BDIO_write(fb,p%data,  .true.)
    call BDIO_write_hash(fb)
    
    
    return
  end subroutine write_bdio

! ****************************************
! *
  module subroutine read_bdio(p, fb)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    type (BDIO), intent (inout) :: fb

    integer :: ir, n, nid
    integer, allocatable :: nd(:), nrep(:), ivrep(:)
    real (kind=DP) :: dfoo(1)
    integer :: ifoo(1)
    type (uwreal) :: temp
    logical :: remember_mode

    remember_mode = light_mode

    ir = 0
    ir = ir + BDIO_read(fb,dfoo)
    ir = ir + BDIO_read(fb,ifoo)
    nid = ifoo(1)
    
    allocate(nd(nid), nrep(nid))
    ir = ir + BDIO_read(fb,nd)
    ir = ir + BDIO_read(fb,nrep)

    n = sum(nrep)
    allocate(ivrep(n))
    ir = ir + BDIO_read(fb,ivrep)

    light_mode = .false.
    call p%init(nid, nd, nrep, ivrep)
    p%mean = dfoo(1)
    ir = ir + BDIO_read(fb,p%id)
    ir = ir + BDIO_read(fb,p%tmax)
    ir = ir + BDIO_read(fb,p%texp)
    ir = ir + BDIO_read(fb,p%s)
    ir = ir + BDIO_read(fb,p%data)

    if (remember_mode) then
       nd    = 1
       nrep  = 1
       ivrep = 1
       
       call temp%init(nid, nd, nrep, ivrep)
       call p%uwerr()
       

       p = temp
    end if
    
    light_mode = remember_mode
    
    return
  end subroutine read_bdio

end submodule aderrors_io

! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:

