!
! PROGRAM simple 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    complete.f90
! created: Tue Aug 21 19:45:22 2018
!                               

program complete

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use constants
  use aderrors
  use simulator
  
  implicit none

  integer, parameter  :: nd = 5000, nrep=4
  type (uwreal)       :: x, y, z
  integer             :: iflog, ivrep(nrep)=(/1000,30,3070,900/), i, is, ie
  real (kind=DP)      :: data_x(nd), data_y(nd/2), err, ti, texp
  real (kind=DP)      :: tau(4), &
       lam_x(4)=(/1.0_DP, 0.70_DP, 0.40_DP, 0.40_DP/), &
       lam_y(4)=(/2.3_DP, 0.40_DP, 0.20_DP, 0.90_DP/)
  character (len=200) :: flog='history_z.log'


  ! Fill arrays data_x(:) with autocorrelated
  ! data from the module simulator. Use nrep replica
  tau = (/1.0_DP, 3.0_DP, 12.0_DP, 75.0_DP/)
  texp = maxval(tau)

  is = 1
  do i = 1, nrep
     ie = is + ivrep(i) - 1
     call gen_series(data_x(is:ie),  err, ti, tau, lam_x, 0.3_DP)
     is = ie + 1
  end do

  ! Fill data_y(:) with different values of tau also using
  ! module simulator
  forall (i=1:4) tau(i) = real(2*i,kind=DP)
  call gen_series(data_y, err, ti, tau, lam_y, 1.3_DP)


  ! Load data_x(:) measurements in variable x.
  ! Set replica vector, exponential autocorrelation time
  ! and ensemble ID.
  x = data_x
  call x%set_id(1)
  call x%set_replica(ivrep)
  call x%set_texp(texp)
  
  ! Load data_y(:) measurements in variable y
  y = data_y
  call y%set_id(2)

  ! Exact, transparent error propagation
  z = sin(x)/(cos(y) + 1.0_DP)

  ! Attach tail in ensemble with ID 1 when signal in the
  ! normalized auto-correlation function equals its error
  call z%set_dsig(1.0_DP,1) 
  ! Set Stau=3 for automatic window in ensemble with ID 2
  call z%set_stau(3.0_DP,2)
  ! Perform error analysis (tails, optimal window,...)
  call z%uwerr() 

  ! Print results and output details to flog  
  write(*,'(1A,1F8.5,1A,1F8.5)')'** Observable z:   ', z%value(),  " +/- ", z%error()
  do i = 1, z%neid()
     write(*,'(3X,1A,1I3,3X,1F5.2,"%")',advance="no")'Contribution to error from ensemble ID', &
          z%eid(i), 100.0_DP*z%error_src(i)
     write(*,'(2X,1A,1F0.4,1A,1F8.4,1A)')'(tau int: ', z%taui(i), " +/- ", z%dtaui(i), ")"
  end do

  open(newunit=iflog, file=trim(flog))
  call z%print_hist(iflog)
  close(iflog)
  
  stop
end program complete

