program newton

  use numtypes
  use aderrors

  implicit none
  
  type (uwreal)  :: xant, xnew, val, der, a, b
  integer        :: id, i
  
  a = (/ 0.8_DP, 0.056_DP /)
  id = aderr_new_id("Variable a", 100)
  call a%set_id(id)

  b = (/ 1.3_DP, 0.10_DP /)
  id = aderr_new_id("Variable b", 222)
  call b%set_id(id)

  xnew = (/0.50_DP, 10.0_DP/)
  id = aderr_new_id("Initial condition", 3674)
  call xnew%set_id(id)

  ! Newton's method. Iterate until root does not change 
  do
    xant = xnew
    call fnew(xant, val, der)
    xnew = xant - val/der
    if (abs(xnew%value()-xant%value()).lt.1.0E-10_DP) exit
  end do

  call xnew%uwerr()
  write(*,*)"Fixed point at: ", &
            xnew%value(), " +/- ", xnew%error()
  do i = 1, xnew%neid()
     write(*,200)trim(xnew%error_name(i)), &
          ' contribution to error: ', xnew%error_src(i)*100., '%'
  end do

200 FORMAT(1A20,1A,5X,1F6.2,1A)
  
  stop          
contains
  subroutine fnew(x, f, d)
  ! Returns the value (f) an derivative (d) of
  ! the function f(x)-x
      
    type (uwreal), intent (in)  :: x
    type (uwreal), intent (out) :: f, d

    f =  a*cos(b*sin(x)) - x
    d = -a*b*sin(b*sin(x)) * cos(x) - 1.0_DP

    return
  end subroutine fnew
end program newton
