!
! PROGRAM newton_grad 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    newton_grad.f90
! created: Mon Sep  3 00:01:06 2018
!                               

program newton_grad

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use aderrors
  
  implicit none

  type (uwreal)  :: pa, pb, xini, a, b
  real (kind=DP) :: grad(3)
  integer        :: id
  
  pa = (/ 0.8_DP, 0.056_DP /)
  id = aderr_new_id("Variable a", 100)
  call pa%set_id(id)

  pb = (/ 1.3_DP, 0.10_DP /)
  id = aderr_new_id("Variable b", 222)
  call pb%set_id(id)

  xini = (/0.50_DP, 10.0_DP/)
  id = aderr_new_id("Initial condition", 3674)
  call xini%set_id(id)

  call aderr_grad(grad, newton, [pa, pb, xini])
  write(*,*)'derivative of newton method respect a:    ', grad(1)
  write(*,*)'derivative of newton method respect b:    ', grad(2)
  write(*,*)'derivative of newton method respect xini: ', grad(3)

  stop
contains

  function newton(x)
    type (uwreal) :: newton
    type (uwreal), intent (in) :: x(:)

    type (uwreal) :: xant, xnew, val, der
    
    a    = x(1)
    b    = x(2)
    xnew = x(3)
    do
       xant = xnew
       call fnew(xant, val, der)
       xnew = xant - val/der
       if (abs(xnew%value()-xant%value()).lt.1.0E-10_DP) exit
    end do

    newton = xnew
    
    return
  end function newton
  
  subroutine fnew(x, f, d)
  ! Returns the value (f) an derivative (d) of
  ! the function f(x)-x
      
    type (uwreal), intent (in)  :: x
    type (uwreal), intent (out) :: f, d

    f =  a*cos(b*sin(x)) - x
    d = -a*b*sin(b*sin(x)) * cos(x) - 1.0_DP

    return
  end subroutine fnew
end program newton_grad

